/* File:        flraggregate_dyn.P: 
**     	        trailer appended automatically by the FLORA compiler
**
** Author(s): Guizhen Yang 
** Contact:   xsb-contact@cs.sunysb.edu
**
** Copyright (C) The Research Foundation of SUNY, 1999
**
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
** 
** $Id: flraggregate_dyn.P,v 1.7 1999/11/21 05:52:38 kifer Exp $
** 
*/

:- compiler_options([xpp_on]).

#define  VAR_FPREFIX_BASE  dyn_$_$_flora_
#include "flora_prefix.h"


:- import
	var/1, functor/3, arg/3, telling/1, tell/1, nl/0, writeln/1, number/1
   from standard.

:- import member/2, length/2 from basics.

:- import findall/3, sort/2 from setof.

:- import STD_FPREFIX(warning_nl)/1 from flrutils.


/****************************************************************************
  rules for aggregates
****************************************************************************/
VAR_FPREFIX(min)(G,Matches,Min) :-
	( G = nogrp ->
	    Matches=[H|T],
	    VAR_FPREFIX(minaux1)(T,H,Min)

	  ; VAR_FPREFIX(var_exists)(G) ->
	    findall(Grp, member(flora_rslt(_,Grp), Matches), L),
	    sort(L,Groups),
	    member(G,Groups),
	    findall(D,member(flora_rslt(D,G),Matches),DL),
	    DL=[H|T],
	    VAR_FPREFIX(minaux1)(T,H,Min)

	  ;
	    Matches=[flora_rslt(D,_)|T],
	    VAR_FPREFIX(minaux2)(T,D,Min)
	).

VAR_FPREFIX(minaux1)([],X,X).
VAR_FPREFIX(minaux1)([H|T],X,N) :-
	(H @< X
	->  VAR_FPREFIX(minaux1)(T,H,N)
	;   VAR_FPREFIX(minaux1)(T,X,N)
	).

VAR_FPREFIX(minaux2)([],X,X).
VAR_FPREFIX(minaux2)([flora_rslt(D,_)|T],X,N) :-
	(D @< X
	->  VAR_FPREFIX(minaux2)(T,D,N)
	;   VAR_FPREFIX(minaux2)(T,X,N)
	).


/****************************************************************************/
VAR_FPREFIX(max)(G,Matches,Max) :-
	( G = nogrp ->
	    Matches=[H|T],
	    VAR_FPREFIX(maxaux1)(T,H,Max)

	  ; VAR_FPREFIX(var_exists)(G) ->
	    findall(Grp,member(flora_rslt(_,Grp),Matches),L),
	    sort(L,Groups),
	    member(G,Groups),
	    findall(D,member(flora_rslt(D,G),Matches),DL),
	    DL=[H|T],
	    VAR_FPREFIX(maxaux1)(T,H,Max)

	  ;
	    Matches=[flora_rslt(D,_)|T],
	    VAR_FPREFIX(maxaux2)(T,D,Max)
	).

VAR_FPREFIX(maxaux1)([],X,X).
VAR_FPREFIX(maxaux1)([H|T],X,N) :-
	(H @> X
	->  VAR_FPREFIX(maxaux1)(T,H,N)
	;   VAR_FPREFIX(maxaux1)(T,X,N)
	).

VAR_FPREFIX(maxaux2)([],X,X).
VAR_FPREFIX(maxaux2)([flora_rslt(D,_)|T],X,N) :-
	(D @> X
	->  VAR_FPREFIX(maxaux2)(T,D,N)
	;   VAR_FPREFIX(maxaux2)(T,X,N)
	).


/****************************************************************************/
VAR_FPREFIX(sum)(G,Matches,Sum) :-
	( G = nogrp ->
	    VAR_FPREFIX(sumaux1)(Matches,X,Sum,Warning)

	  ; VAR_FPREFIX(var_exists)(G) ->
	    findall(Grp,member(flora_rslt(_,Grp),Matches),L),
	    sort(L,Groups),
	    member(G,Groups),
	    findall(D, member(flora_rslt(D,G), Matches), DL),
	    VAR_FPREFIX(sumaux1)(DL,X,Sum,Warning)

	  ;
	    VAR_FPREFIX(sumaux2)(Matches,X,Sum,Warning), !
	),
	( var(Warning) ->
	    true
	  ;
	    STD_FPREFIX(warning_nl)('discarding non-numeric objects while computing sum')
	).

VAR_FPREFIX(sumaux1)([],X,X,_) :-
	(var(X) -> fail ; true).

VAR_FPREFIX(sumaux1)([H|T],X,N,Warning) :-
	( number(H) ->
	    (var(X) -> S=H ; S is X+H),
	    VAR_FPREFIX(sumaux1)(T,S,N,Warning)

	  ;
	    Warning=warning,
	    VAR_FPREFIX(sumaux1)(T,X,N,_)
	).

VAR_FPREFIX(sumaux2)([],X,X,_) :-
	(var(X) -> fail ; true).

VAR_FPREFIX(sumaux2)([flora_rslt(D,_)|T],X,N,Warning) :-
	( number(D) ->
	    (var(X) -> S=D ; S is X+D),
	    VAR_FPREFIX(sumaux2)(T,S,N,Warning)

	  ;
	    Warning=warning,
	    VAR_FPREFIX(sumaux2)(T,X,N,_)
	).


/****************************************************************************/
VAR_FPREFIX(avg)(G,Matches,Avg) :-
	( G = nogrp ->
	    VAR_FPREFIX(avgaux1)(Matches,X1,X2,Avg,Warning)

	  ; VAR_FPREFIX(var_exists)(G) ->
	    findall(Grp,member(flora_rslt(_,Grp),Matches),L),
	    sort(L,Groups),
	    member(G,Groups),
	    findall(D,member(flora_rslt(D,G),Matches),DL),
	    VAR_FPREFIX(avgaux1)(DL,X1,X2,Avg,Warning)

	  ;
	    VAR_FPREFIX(avgaux2)(L,X1,X2,Avg,Warning)
	),
	( var(Warning) ->
	    true
	  ;
	    STD_FPREFIX(warning_nl)('discarding non-numeric objects while computing average')
	).

VAR_FPREFIX(avgaux1)([],S,N,Avg,_) :-
	(var(S) -> fail ; Avg is S/N).

VAR_FPREFIX(avgaux1)([H|T],X1,X2,Avg,Warning) :-
	( number(H) ->
	    (var(X1) -> S=H, N=1 ; S is X1+H, N is X2+1),
	    VAR_FPREFIX(avgaux1)(T,S,N,Avg,Warning)

	  ;
	    Warning=warning,
	    VAR_FPREFIX(avgaux1)(T,X1,X2,Avg,_)
	).

VAR_FPREFIX(avgaux2)([],S,N,Avg,_) :-
	(var(S) -> fail ; Avg is S/N).

VAR_FPREFIX(avgaux2)([flora_rslt(D,_)|T],X1,X2,Avg,Warning) :-
	( number(D) ->
	    (var(X1) -> S=D, N=1 ; S is X1+D, N is X2+1),
	    VAR_FPREFIX(avgaux2)(T,S,N,Avg,Warning)

	  ;
	    Warning=warning,
	    VAR_FPREFIX(avgaux2)(T,X1,X2,Avg,_)
	).


/****************************************************************************/
VAR_FPREFIX(count)(G,Matches,Count) :-
	( VAR_FPREFIX(var_exists)(G) ->
	    findall(Grp,member(flora_rslt(_,Grp),Matches),L),
	    sort(L,Groups),
	    member(G,Groups),
	    findall(D,member(flora_rslt(D,G),Matches),DL),
	    length(DL,Count)

	  ;
	    length(Matches,Count)
	).


/****************************************************************************/
VAR_FPREFIX(collectset)(G,Matches,Set) :-
	( G = nogrp ->
	    sort(Matches,Set)

	  ; VAR_FPREFIX(var_exists)(G) ->
	    findall(Grp,member(flora_rslt(_,Grp),Matches),L),
	    sort(L,Groups),
	    member(G,Groups),
	    findall(D,member(flora_rslt(D,G),Matches),DL),
	    sort(DL,Set)

	  ;
	    findall(D,member(flora_rslt(D,G),Matches),DL),
	    sort(DL,Set)
	).


/****************************************************************************/
VAR_FPREFIX(collectbag)(G,Matches,Bag) :-
	( G = nogrp ->
	    Bag=Matches

	  ; VAR_FPREFIX(var_exists)(G) ->
	    findall(Grp,member(flora_rslt(_,Grp),Matches),L),
	    sort(L,Groups),
	    member(G,Groups),
	    findall(D,member(flora_rslt(D,G),Matches),Bag)

	  ;
	    findall(D,member(flora_rslt(D,G),Matches),Bag)
	).


/****************************************************************************/
VAR_FPREFIX(var_exists)(X) :- var(X), !.

VAR_FPREFIX(var_exists)(T) :-
	functor(T,_,N),
	VAR_FPREFIX(var_exists)(T,N).

VAR_FPREFIX(var_exists)(T,N) :-
	N>=1,
	arg(N,T,Arg),
	( VAR_FPREFIX(var_exists)(Arg) ->
	    !
	  ;
	    M is N-1,
	    VAR_FPREFIX(var_exists)(T,M)
	).


#undef VAR_FPREFIX_BASE

/*******************  End of flraggregate_dyn.P  ********************/

